/* Metageek WiSPY interface 
 * Mike Kershaw/Dragorn <dragorn@kismetwireless.net>
 *
 * This code is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This code is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * Extra thanks to Ryan Woodings @ Metageek for interface documentation
 */

#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <gtk/gtk.h>
#include <gtk/gtkmain.h>
#include <gtk/gtksignal.h>

#include "wispy_gtk_widget.h"

/* control/picker pane width and initial height */
#define WISPY_WIDGET_PADDING	5

static void wispy_widget_class_init(WispyWidgetClass *class);
static void wispy_widget_init(WispyWidget *graph);
static void wispy_widget_destroy(GtkObject *object);
static void wispy_widget_realize(GtkWidget *widget);

static gint wispy_widget_configure(GtkWidget *widget, 
									 GdkEventConfigure *event);
static gboolean wispy_widget_expose(GtkWidget *widget, 
									  GdkEventExpose *event,
									  gpointer *aux);
static gint wispy_widget_timeout(gpointer *data);
static void wispy_widget_size_allocate(GtkWidget *widget,
										 GtkAllocation *allocation);
static void wispy_widget_wdr_sweep(int slot, int mode,
								   wispy_sample_sweep *sweep, void *aux);

G_DEFINE_TYPE(WispyWidget, wispy_widget, GTK_TYPE_EVENT_BOX);
						
static void wispy_widget_class_init(WispyWidgetClass *class) {
	GObjectClass *gobject_class;
	GtkObjectClass *object_class;
	GtkWidgetClass *widget_class;

	gobject_class = G_OBJECT_CLASS(class);
	object_class = GTK_OBJECT_CLASS(class);
	widget_class = GTK_WIDGET_CLASS(class);

	object_class->destroy = wispy_widget_destroy;
	widget_class->size_allocate = wispy_widget_size_allocate;
}

static void wispy_widget_destroy(GtkObject *object) {
	WispyWidget *wwidget = WISPY_WIDGET(object);

	wdr_del_sweepcb(wwidget->wdr, wwidget->wdr_slot,
					wispy_widget_wdr_sweep, wwidget);

	if (wwidget->wdr_slot >= 0) {
		wdr_del_ref(wwidget->wdr, wwidget->wdr_slot);
		wwidget->wdr_slot = -1;
	}

	if (wwidget->timeout_ref >= 0) {
		g_source_remove(wwidget->timeout_ref);
		wwidget->timeout_ref = -1;
	}

	GTK_OBJECT_CLASS(wispy_widget_parent_class)->destroy(object);
}

GtkWidget *wispy_widget_new(wispy_device_registry *wdr) {
	WispyWidget *wwidget;

	wwidget = gtk_type_new(wispy_widget_get_type());

	wwidget->wdr = wdr;

	return GTK_WIDGET(wwidget);
}

inline int wispy_widget_find_chan_pt(WispyWidget *wwidget, int x, int y) {
	int dbm, maxy, nchannels;

	/* Only compute if we know a chanset and if we're inside the bounding
	 * box */
	if (wwidget->chanset == NULL)
		return -2;

	if (x < wwidget->chan_start_x || x > wwidget->chan_end_x ||
		y < wwidget->chan_start_y || y > wwidget->chan_end_y) {
		return -2;
	}

	/* Search for the channel positions */
	maxy = 0;
	nchannels = wwidget->chanset->chan_num;
	for (dbm = 0; dbm < nchannels; dbm++) {
		if (wwidget->chan_points[dbm + nchannels].y > maxy)
			maxy = wwidget->chan_points[dbm + nchannels].y;

		if (x > wwidget->chan_points[dbm].x && 
			x < wwidget->chan_points[dbm + nchannels].x &&
			y > wwidget->chan_points[dbm].y &&
			y < wwidget->chan_points[dbm + nchannels].y) {

			return dbm;
		}
	}

	return -1;
}

static void wispy_widget_wdr_sweep(int slot, int mode,
								   wispy_sample_sweep *sweep, void *aux) {
	WispyWidget *wwidget;

	g_return_if_fail(aux != NULL);
	g_return_if_fail(IS_WISPY_WIDGET(aux));

	wwidget = WISPY_WIDGET(aux);

	/* Generic sweep handler to add it to our cache, all things get this */
	if ((mode & WISPY_POLL_ERROR)) {
		wwidget->phydev = NULL;
		wispy_cache_free(wwidget->sweepcache);
		wwidget->sweepcache = NULL;
		wdr_del_ref(wwidget->wdr, wwidget->wdr_slot);
		wwidget->wdr_slot = -1;
	} else if ((mode & WISPY_POLL_CONFIGURED)) {
		/* TODO - multi-sweep-range support */
		wwidget->base_db_offset = 
			wwidget->phydev->device_spec->supported_ranges[0].max_sample;
		wwidget->min_db_draw = 
			wwidget->phydev->device_spec->supported_ranges[0].min_sample;
	} else if (wwidget->sweepcache != NULL && sweep != NULL) {
		wispy_cache_append(wwidget->sweepcache, sweep);
	}

	/* Call the secondary sweep handler */
	if (wwidget->wdr_sweep_func != NULL)
		(*(wwidget->wdr_sweep_func))(slot, mode, sweep, aux);
}

/* Common level function for opening a device, calls the secondary level
 * function if one exists after device is linked in */
static void wispy_widget_wdr_devopen(int slot, void *aux) {
	WispyWidget *wwidget;
	char ltxt[256];
	wispy_sample_sweep *ran;

	g_return_if_fail(aux != NULL);
	g_return_if_fail(IS_WISPY_WIDGET(aux));

	wwidget = WISPY_WIDGET(aux);

	/* Unref using the old slot, but only after we've reffed using the 
	 * new one, incase someone picked the same dev twice, we need this ordering
	 * to prevent it from getting reaped from a 0 count on the ref*/
	wdr_add_ref(wwidget->wdr, slot);
	wdr_del_ref(wwidget->wdr, wwidget->wdr_slot);

	/* Allocate the new device and drop our old sweep cache if one existed */
	wwidget->wdr_slot = slot;
	wwidget->phydev = wdr_get_phy(wwidget->wdr, slot);

	if (wwidget->sweepcache != NULL)
		wispy_cache_free(wwidget->sweepcache);

	wwidget->chanset = NULL;
	if (wwidget->chanhit)
		free(wwidget->chanhit);
	wwidget->chanhit = NULL;
	if (wwidget->chancolors)
		free(wwidget->chancolors);
	wwidget->chancolors = NULL;

	/* Allocate using our options */
	wwidget->sweepcache = 
		wispy_cache_alloc(wwidget->sweep_num_samples, 
						  wwidget->sweep_keep_avg, 
						  wwidget->sweep_keep_peak);

	/* register a sweep callback */
	wdr_add_sweepcb(wwidget->wdr, wwidget->wdr_slot,
					wispy_widget_wdr_sweep, wwidget->sweep_num_aggregate,
					wwidget);

	/* Call our secondary open function */
	if (wwidget->wdr_devopen_func != NULL)
		(*(wwidget->wdr_devopen_func))(slot, aux);

	/* Force calibration */
	if (wispy_get_state(wwidget->phydev) > WISPY_STATE_CONFIGURING)
		wispy_widget_wdr_sweep(-1, WISPY_POLL_CONFIGURED, NULL, aux);
}

/* Picks an existing device from the dynamic WDR list in the 
 * popup menu */
static void wispy_widget_menu_devpicker(gpointer *aux) {
	WispyWidget *wwidget;
	wdr_menu_rec *r;

	g_return_if_fail(aux != NULL);

	/* We're opening a known device, call it directly */
	r = (wdr_menu_rec *) aux;
	wwidget = WISPY_WIDGET(r->aux);

	if (r->slot < 0) {
		wdr_devpicker_spawn(wwidget->wdr,
							wispy_widget_wdr_devopen,
							wwidget);
	}

	wispy_widget_wdr_devopen(r->slot, wwidget);
}

/* Spawns the device picker window from the popup menu */
static void wispy_widget_menu_spawnpicker(gpointer *aux) {
	WispyWidget *wwidget;

	g_return_if_fail(aux != NULL);
	g_return_if_fail(IS_WISPY_WIDGET(aux));

	wwidget = WISPY_WIDGET(aux);

	wdr_devpicker_spawn(wwidget->wdr,
						wispy_widget_wdr_devopen,
						wwidget);
}

/* Spawns the network picker window from the popup menu */
static void wispy_widget_menu_spawnnetpicker(gpointer *aux) {
	WispyWidget *wwidget;

	g_return_if_fail(aux != NULL);
	g_return_if_fail(IS_WISPY_WIDGET(aux));

	wwidget = WISPY_WIDGET(aux);

	wdr_netpicker_spawn(wwidget->wdr,
						wispy_widget_wdr_devopen,
						wwidget);
}

/* Called when the popup button menu goes away, responsible for freeing the
 * dynamic wdr menu items (which in turn decrements the use counter so we can
 * release devices if we need to) */
static void wispy_widget_menu_destroy(gpointer *aux) {
	WispyWidget *wwidget;

	g_return_if_fail(aux != NULL);
	g_return_if_fail(IS_WISPY_WIDGET(aux));

	wwidget = WISPY_WIDGET(aux);

	wdr_free_menu(wwidget->wdr, wwidget->wdr_menu);
}

static gboolean wispy_widget_menu_button_press(GtkWidget *widget,
											   GdkEvent *event) {
	GtkWidget *menu, *mi;
	GList *iter;
	WispyWidget *wwidget;

	g_return_if_fail(widget != NULL);
	g_return_if_fail(IS_WISPY_WIDGET(widget));
	g_return_if_fail(event != NULL);

	wwidget = WISPY_WIDGET(widget);

	if (event->type == GDK_BUTTON_PRESS) {
		GdkEventButton *bevent = (GdkEventButton *) event;

		menu = gtk_menu_new();
		gtk_widget_show(menu);

		if (wwidget->wdr_slot < 0) {
			mi = gtk_menu_item_new_with_label("No current device...");
		} else {
			mi = gtk_menu_item_new_with_label(wispy_phy_getname(wwidget->phydev));
		}
		gtk_widget_set_sensitive(mi, 0);
		gtk_menu_shell_append(GTK_MENU_SHELL(menu), mi);
		gtk_widget_show(mi);

		mi = gtk_separator_menu_item_new();
		gtk_menu_shell_append(GTK_MENU_SHELL(menu), mi);
		gtk_widget_show(mi);

		/* Add the WDR generated menus, if any */
		wwidget->wdr_menu = 
			wdr_populate_menu(wwidget->wdr,
							  GTK_WIDGET(menu),
							  G_CALLBACK(wispy_widget_menu_devpicker),
							  wwidget);

		mi = gtk_menu_item_new_with_label("Open Device...");
		gtk_menu_shell_append(GTK_MENU_SHELL(menu), mi);
	    g_signal_connect_swapped(G_OBJECT(mi), "activate",
								 G_CALLBACK(wispy_widget_menu_spawnpicker), 
								 wwidget);
		gtk_widget_show(mi);

		mi = gtk_menu_item_new_with_label("Open Network Device...");
		gtk_menu_shell_append(GTK_MENU_SHELL(menu), mi);
	    g_signal_connect_swapped(G_OBJECT(mi), "activate",
								 G_CALLBACK(wispy_widget_menu_spawnnetpicker), 
								 wwidget);
		gtk_widget_show(mi);

		/* Add the end of the menu and the close option */
		mi = gtk_separator_menu_item_new();
		gtk_menu_shell_append(GTK_MENU_SHELL(menu), mi);
		gtk_widget_show(mi);

		mi = gtk_menu_item_new_with_label("Close graph");
		gtk_menu_shell_append(GTK_MENU_SHELL(menu), mi);
		gtk_widget_show(mi);
		g_signal_connect_swapped(G_OBJECT(mi), "activate",
								 G_CALLBACK(gtk_widget_destroy), G_OBJECT(widget));

		/* Set the cleanup function for the dynamic menu generation */
		g_signal_connect_swapped(G_OBJECT(menu), "selection-done",
								 G_CALLBACK(wispy_widget_menu_destroy),
								 wwidget);

		gtk_menu_popup(GTK_MENU(menu), NULL, NULL, NULL, NULL, 
					   bevent->button, bevent->time);

		return TRUE;
	}

	return FALSE;
}

void wispy_widget_buildgui(WispyWidget *widget) {
	GtkWidget *arrow, *button2, *label;
	GtkWidget *vbox, *hbox, *hbox2;
	/* Sigh, is this really the only way to change the colors here? */
	GtkWidget *eb;
	
	GtkWidget *closebutton;

	GdkColor c;
	GtkStyle *style;

	g_return_if_fail(widget != NULL);

	/* Brute override of the style */
	gdk_color_parse("#505050", &c);
	style = gtk_style_new();
	gtk_widget_set_style(GTK_WIDGET(widget), style);
	style->bg[GTK_STATE_NORMAL] = c;
	style->base[GTK_STATE_NORMAL] = c;
	gtk_style_unref(style);

	/* Main packing of graph and sidebar */
	hbox = gtk_hbox_new(FALSE, WISPY_WIDGET_PADDING);
	gtk_container_add(GTK_CONTAINER(widget), hbox);
	gtk_widget_show(hbox);

	/* Make the cairo drawing area and link it into events */
	widget->draw = gtk_drawing_area_new();
	gtk_signal_connect(GTK_OBJECT(widget->draw), "expose_event",
					   (GtkSignalFunc) wispy_widget_expose, widget);

	/* Set mask */
	gtk_widget_set_events(widget->draw, GDK_EXPOSURE_MASK |
						  GDK_LEAVE_NOTIFY_MASK |
						  GDK_BUTTON_PRESS_MASK |
						  GDK_POINTER_MOTION_MASK |
						  GDK_POINTER_MOTION_HINT_MASK);

	/* Attach mouse */
	if (widget->draw_mouse_click_func != NULL) 
		gtk_signal_connect(GTK_OBJECT(widget->draw), "button_press_event",
						   (GtkSignalFunc) widget->draw_mouse_click_func, 
						   widget);
	if (widget->draw_mouse_move_func != NULL)
		gtk_signal_connect(GTK_OBJECT(widget->draw), "motion_notify_event",
						   (GtkSignalFunc) widget->draw_mouse_move_func, 
						   widget);

	gtk_widget_show(widget->draw);
	gtk_box_pack_start(GTK_BOX(hbox), widget->draw, TRUE, TRUE, 0);

	/* Sidebar contents */
	vbox = gtk_vbox_new(FALSE, 0);
	gtk_container_set_border_width(GTK_CONTAINER(vbox), WISPY_WIDGET_PADDING);
	gtk_box_pack_end(GTK_BOX(hbox), vbox, FALSE, FALSE, 0);
	gtk_widget_show(vbox);

	/* Colored titled box and label */
	eb = gtk_event_box_new();
	gtk_box_pack_start(GTK_BOX(vbox), eb, FALSE, FALSE, 0);

	if (widget->graph_title_bg != NULL) {
		gdk_color_parse(widget->graph_title_bg, &c);
		style = gtk_style_new();
		gtk_widget_set_style(GTK_WIDGET(eb), style);
		style->bg[GTK_STATE_NORMAL] = c;
		gtk_style_unref(style);
	} else {
		fprintf(stderr, "BUG: Missing graph_title_bg in widget base\n");
	}

	if (widget->graph_title != NULL) {
		hbox2 = gtk_hbox_new(FALSE, 2);
		gtk_container_add(GTK_CONTAINER(eb), hbox2);
		gtk_widget_show(hbox2);

		label = gtk_label_new(NULL);
		gtk_label_set_markup(GTK_LABEL(label), widget->graph_title);
		gtk_box_pack_start(GTK_BOX(hbox2), label, FALSE, FALSE, 0);

		arrow = gtk_arrow_new(GTK_ARROW_DOWN, GTK_SHADOW_OUT);
		gtk_widget_show(arrow);
		widget->menubutton = gtk_button_new();
		gtk_container_add(GTK_CONTAINER(widget->menubutton), arrow);
		gtk_box_pack_end(GTK_BOX(hbox2), widget->menubutton, FALSE, FALSE, 0);
		g_signal_connect_swapped(G_OBJECT(widget->menubutton), "event", 
								 G_CALLBACK(wispy_widget_menu_button_press),
								 G_OBJECT(widget));
		gtk_widget_show(widget->menubutton);


		gtk_widget_show(eb);
		gtk_widget_show(label);
	} else {
		fprintf(stderr, "BUG: Missing graph_title in widget base\n");
	}

	eb = gtk_event_box_new();
	gtk_box_pack_start(GTK_BOX(vbox), eb, TRUE, TRUE, 0);

	if (widget->graph_control_bg != NULL) {
		gdk_color_parse(widget->graph_control_bg, &c);
		style = gtk_style_new();
		gtk_widget_set_style(GTK_WIDGET(eb), style);
		style->bg[GTK_STATE_NORMAL] = c;
		gtk_style_unref(style);
	} else {
		fprintf(stderr, "BUG: MIssing graph_control_bg in widget base\n");
	}

	widget->vbox = gtk_vbox_new(FALSE, 0);
	gtk_container_add(GTK_CONTAINER(eb), widget->vbox);
	gtk_widget_show(widget->vbox);
	gtk_widget_show(eb);

#if 0
	closebutton = gtk_button_new_with_label("Close Graph");
	g_signal_connect_swapped(G_OBJECT(closebutton), "clicked",
							 G_CALLBACK(gtk_widget_destroy), G_OBJECT(widget));
	gtk_box_pack_end(GTK_BOX(widget->vbox), closebutton, FALSE, FALSE, 0);
	gtk_widget_show(closebutton);
#endif
}

static void wispy_widget_init(WispyWidget *widget) {
	widget->hi_chan = -1;
	widget->g_start_x = widget->g_end_x = widget->g_len_x = 0;
	widget->g_start_y = widget->g_end_y = widget->g_len_y = 0;
	widget->chan_points = NULL;
	widget->chan_start_x = widget->chan_end_x = 
		widget->chan_start_y = widget->chan_end_y = 0;
	widget->chanset = NULL;
	widget->chanhit = NULL;
	widget->chancolors = NULL;

	widget->timeout_ref = 
		g_timeout_add(500, (GSourceFunc) wispy_widget_timeout, widget);

	widget->phydev = NULL;
	widget->sweepcache = NULL;
	widget->wdr_menu = NULL;
	widget->wdr_slot = -1;
}

static void wispy_widget_size_allocate(GtkWidget *widget,
										 GtkAllocation *allocation) {
	WispyWidget *wwidget;

	g_return_if_fail(IS_WISPY_WIDGET(widget));
	g_return_if_fail(allocation != NULL);

	wwidget = WISPY_WIDGET(widget);

	/* Size our control panel as a fixed size 
	 * TODO - revisit this and maybe have a minimum size as well? */
	gtk_widget_set_size_request(wwidget->vbox, allocation->width / 5 + 20, -1);

	GTK_WIDGET_CLASS(wispy_widget_parent_class)->size_allocate(widget, allocation);
}

void wispy_widget_draw(GtkWidget *widget, cairo_t *cr, WispyWidget *wwidget) {
	cairo_text_extents_t extents;
	int x, chpix, maxcw;
	const double dash_onoff[] = {2, 4};
	const double dash_ononoff[] = {4, 2};

	char mtext[128];

	int chanmod;

	g_return_if_fail(widget != NULL);
	g_return_if_fail(widget != NULL);

	cairo_save(cr);
	cairo_rectangle(cr, 0, 0, widget->allocation.width, widget->allocation.height);
	cairo_set_source_rgb(cr, HC2CC(0x50), HC2CC(0x50), HC2CC(0x50));
	cairo_fill(cr);
	cairo_stroke(cr);

	wwidget->g_len_x = widget->allocation.width - (WISPY_WIDGET_PADDING * 2);
	wwidget->g_len_y = widget->allocation.height - (WISPY_WIDGET_PADDING * 2);
	wwidget->g_start_x = WISPY_WIDGET_PADDING;
	wwidget->g_start_y = WISPY_WIDGET_PADDING;
	wwidget->g_end_x = wwidget->g_start_x + wwidget->g_len_x;
	wwidget->g_end_y = wwidget->g_start_y + wwidget->g_len_y;

	/* We haven't been initialized so we don't know our channels */
	if (wwidget->wdr_slot < 0) {
		cairo_set_source_rgb(cr, HC2CC(0xFF), HC2CC(0xFF), HC2CC(0xFF));
		cairo_select_font_face(cr, "Helvetica", 
							   CAIRO_FONT_SLANT_NORMAL, 
							   CAIRO_FONT_WEIGHT_BOLD);
		cairo_set_font_size(cr, 16);
		cairo_text_extents(cr, "No device, open one from the graph menu", &extents);
		cairo_move_to(cr, 
					  wwidget->g_start_x + (wwidget->g_len_x / 2) - 
					  (extents.width / 2),
					  wwidget->g_start_y + (wwidget->g_len_y / 2) - 
					  (extents.height / 2));
		cairo_show_text(cr, "No device, open one from the graph menu");
		cairo_restore(cr);
		return;
	}

	/* We haven't calibrated, so we don't know our channels, etc */
	if (wwidget->sweepcache->pos < 0) {
		cairo_set_source_rgb(cr, HC2CC(0xFF), HC2CC(0xFF), HC2CC(0xFF));
		cairo_select_font_face(cr, "Helvetica", 
							   CAIRO_FONT_SLANT_NORMAL, 
							   CAIRO_FONT_WEIGHT_BOLD);
		cairo_set_font_size(cr, 16);
		cairo_text_extents(cr, "Device calibrating...", &extents);
		cairo_move_to(cr, wwidget->g_start_x + (wwidget->g_len_x / 2) - 
					  (extents.width / 2),
					  wwidget->g_start_y + (wwidget->g_len_y / 2) - 
					  (extents.height / 2));
		cairo_show_text(cr, "Device calibrating...");
		cairo_restore(cr);
		return;
	}

	/* Try to figure out the channels we use for this spectrum */
	for (x = 0; channel_list[x].name != NULL && wwidget->chanset == NULL; x++) {
		if (channel_list[x].startkhz >= 
			wwidget->sweepcache->latest->start_khz &&
			channel_list[x].endkhz <= wwidget->sweepcache->latest->end_khz) {
			int cpos;
			double r, g, b;

			wwidget->chanset = &(channel_list[x]);

			/* Allocate the channels that are 'hit' or highlighted */
			if (wwidget->chanhit)
				free(wwidget->chanhit);
			wwidget->chanhit = malloc(sizeof(int) * wwidget->chanset->chan_num);
			memset(wwidget->chanhit, 0, sizeof(int) * wwidget->chanset->chan_num);

			/* Allocate color sweep */
			if (wwidget->chancolors)
				free(wwidget->chancolors);
			wwidget->chancolors = malloc(sizeof(double) *
										 wwidget->chanset->chan_num * 3);

			for (cpos = 0; cpos < wwidget->chanset->chan_num; cpos++) {
				/* Get the RGB values of a full-intensity color somewhere
				 * along the H slider derived from the channel position */
				hsv_to_rgb(&r, &g, &b, 
						   (360 / wwidget->chanset->chan_num) * cpos, 1, 1);
				/* Convert the hex colors to cairo colors */
				wwidget->chancolors[(3 * cpos) + 0] = HC2CC(r);
				wwidget->chancolors[(3 * cpos) + 1] = HC2CC(g);
				wwidget->chancolors[(3 * cpos) + 2] = HC2CC(b);
			}

			break;
		}
	}

	/* Assume we have at most a 3 digit DBM rating, and figure out the scaling.
	 * We use 000 since they're nice fat digits even with a variable-width font,
	 * and then we slap another 10% on the result. 
	 * We re-use the same string for height calcs for the channel list, just
	 * new size */
	cairo_save(cr);
	snprintf(mtext, 128, "-000 dBm");
	cairo_select_font_face(cr, "Helvetica", 
						   CAIRO_FONT_SLANT_NORMAL, 
						   CAIRO_FONT_WEIGHT_BOLD);
	cairo_set_font_size(cr, 10);
	cairo_text_extents(cr, mtext, &extents);
	wwidget->dbm_w = extents.width + ((double) extents.width * 0.1);
	cairo_set_font_size(cr, 14);
	cairo_text_extents(cr, mtext, &extents);
	wwidget->chan_h = extents.height + ((double) extents.height * 0.1) + 5;
	cairo_restore(cr);


	/* Figure out a square scaled to the number of samples we have and
	 * build the size of the animated graph.  wbar ends up being the 
	 * width of a sample, so we can use that for all sorts of math later */
	wwidget->wbar = 
		(double) (widget->allocation.width - wwidget->dbm_w -
		 (WISPY_WIDGET_PADDING * 2) - 5) / 
		(double) (wwidget->sweepcache->latest->num_samples - 1);
	wwidget->g_len_x = wwidget->wbar * 
		(wwidget->sweepcache->latest->num_samples - 1);
	wwidget->g_len_y = widget->allocation.height - (WISPY_WIDGET_PADDING * 2);
	if (wwidget->show_channels)
		wwidget->g_len_y -= wwidget->chan_h;
	wwidget->g_start_x = WISPY_WIDGET_PADDING + wwidget->dbm_w;
	wwidget->g_start_y = WISPY_WIDGET_PADDING;
	wwidget->g_end_x = wwidget->g_start_x + wwidget->g_len_x;
	wwidget->g_end_y = wwidget->g_start_y + wwidget->g_len_y;

	/* Plot the highlighted center rectangle */
	cairo_rectangle(cr, wwidget->g_start_x, wwidget->g_start_y, 
					wwidget->g_len_x, wwidget->g_len_y);
	cairo_set_source_rgb(cr, HC2CC(0x00), HC2CC(0x00), HC2CC(0x00));
	cairo_fill_preserve(cr);
	cairo_stroke(cr);

	/* Plot the channels if we know how */
	if (wwidget->chanset != NULL && wwidget->show_channels) {
		/* Allocate the channel point array if we haven't yet, so the mouse
		 * handlers know where we've clicked */
		if (wwidget->chan_points == NULL) {
			wwidget->chan_points =
				(GdkPoint *) malloc(sizeof(GdkPoint) *
									wwidget->chanset->chan_num * 2);
		}

		/* Draw the channel text along the bottom */
		cairo_save(cr);
		for (x = 0; x < wwidget->chanset->chan_num; x++) {
			chpix = ((float) wwidget->g_len_x /
					 (wwidget->sweepcache->latest->end_khz -
					  wwidget->sweepcache->latest->start_khz)) *
				(wwidget->chanset->chan_freqs[x] - 
				 wwidget->sweepcache->latest->start_khz);

			if (x == wwidget->hi_chan) {
				cairo_set_source_rgb(cr, HC2CC(0xFF), HC2CC(0xF6), HC2CC(0x00));
				snprintf(mtext, 128, "%s", 
						 wwidget->chanset->chan_text[x]);
			} else {
				cairo_set_source_rgb(cr, 1, 1, 1);
				snprintf(mtext, 128, "%s", 
						 wwidget->chanset->chan_text[x]);
			}

			cairo_move_to(cr, wwidget->g_start_x + chpix, 
						  wwidget->g_end_y);
			cairo_line_to(cr, wwidget->g_start_x + chpix, 
						  wwidget->g_end_y + 3);
			cairo_stroke(cr);

			cairo_select_font_face(cr, "Helvetica", 
								   CAIRO_FONT_SLANT_NORMAL, 
								   CAIRO_FONT_WEIGHT_BOLD);
			cairo_set_font_size(cr, 14);
			cairo_text_extents(cr, mtext, &extents);
			cairo_move_to(cr, 
						  wwidget->g_start_x + chpix - (extents.width / 2),
						  wwidget->g_end_y + 10 + extents.height);
			cairo_show_text(cr, mtext);

			wwidget->chan_points[x].x = 
				wwidget->g_start_x + chpix - (extents.width / 2) - 4;
			wwidget->chan_points[x].y = wwidget->g_end_y + 10 - 4;
			wwidget->chan_points[x + wwidget->chanset->chan_num].x =
				wwidget->chan_points[x].x + extents.width + 8;
			wwidget->chan_points[x + wwidget->chanset->chan_num].y =
				wwidget->chan_points[x].y + extents.height + 10;

			if (wwidget->chanhit[x]) {
				cairo_save(cr);
				cairo_set_source_rgb(cr,
									 wwidget->chancolors[(3 * x) + 0],
									 wwidget->chancolors[(3 * x) + 1],
									 wwidget->chancolors[(3 * x) + 2]);
				cairo_rectangle(cr,
						wwidget->g_start_x + chpix - (extents.width / 2) - 3.5,
						wwidget->g_end_y + 10 - 3.5,
								extents.width + 8,
								extents.height + 10);
				cairo_stroke(cr);
				cairo_restore(cr);

			}
		}

		wwidget->chan_start_x = wwidget->chan_points[0].x - 1;
		wwidget->chan_end_x = 
			wwidget->chan_points[(wwidget->chanset->chan_num * 2) - 1].x + 1;
		wwidget->chan_start_y = wwidget->chan_points[0].y - 1;
		wwidget->chan_end_y = 
			wwidget->chan_points[(wwidget->chanset->chan_num * 2) - 1].y + 1;

		cairo_restore(cr);

	}

	/* Call the second-level draw */
	if (wwidget->draw_func != NULL)
		(*(wwidget->draw_func))(widget, cr, wwidget);

	/* Render the selected channel text on top of any other data */
	if (wwidget->chanset != NULL && (wwidget->show_dbm || wwidget->show_dbm_lines)) {
		/* Draw the dBm lines and power labels */
		maxcw = 0;
		cairo_save(cr);
		cairo_set_line_width(cr, 1);
		for (x = 1; x <= (wwidget->hlines); x++) {
			int dbm, py;
			dbm = wwidget->base_db_offset - 
				(x * ((abs(wwidget->min_db_draw) - abs(wwidget->base_db_offset)) / 
							wwidget->hlines));
			py = (float) wwidget->g_len_y * 
				(float) ((float) (abs(dbm) + wwidget->base_db_offset) /
						 (float) (abs(wwidget->sweepcache->latest->min_sample) +
								  wwidget->base_db_offset));

			if (wwidget->show_dbm_lines) {
				cairo_set_source_rgb(cr, 1, 1, 1);
				cairo_set_dash(cr, dash_onoff, 2, 0);
				/* .5 hack for pixel alignment */
				cairo_move_to(cr, wwidget->g_start_x, 
							  wwidget->g_start_y + py + 0.5);
				cairo_line_to(cr, wwidget->g_end_x, wwidget->g_start_y + py + 0.5);
				cairo_stroke(cr);
			}

			if (wwidget->show_dbm) {
				snprintf(mtext, 128, "%d dBm", dbm);

				cairo_select_font_face(cr, "Helvetica", 
									   CAIRO_FONT_SLANT_NORMAL, 
									   CAIRO_FONT_WEIGHT_BOLD);
				cairo_set_font_size(cr, 10);
				cairo_text_extents(cr, mtext, &extents);
				cairo_move_to(cr, wwidget->g_start_x - wwidget->dbm_w, 
							  wwidget->g_start_y + py + (extents.height / 2));

				cairo_show_text(cr, mtext);
			}

		}
		cairo_restore(cr);

		/* Plot the highlighted channels */
		for (x = 0; x < wwidget->chanset->chan_num; x++) {
			int center, spread;
			int start, end;

			if (wwidget->chanhit[x] == 0 &&
				wwidget->hi_chan != x)
				continue;

			/*
			 * We'll draw the sidebar lines on this at the end so it's on
			 * top of the spectral data
			 */
			center = ((float) wwidget->g_len_x /
					  (wwidget->sweepcache->latest->end_khz -
					   wwidget->sweepcache->latest->start_khz)) *
				(wwidget->chanset->chan_freqs[x] - 
				 wwidget->sweepcache->latest->start_khz) + wwidget->g_start_x;

			spread = ((float) wwidget->g_len_x /
					  (wwidget->sweepcache->latest->end_khz -
					   wwidget->sweepcache->latest->start_khz)) *
				(wwidget->chanset->chan_width); 

			start = center - (spread / 2);
			end = center + (spread / 2);

			if (start < wwidget->g_start_x) {
				start = wwidget->g_start_x;
			}

			if (end > wwidget->g_end_x) {
				end = wwidget->g_end_x;
			}

			cairo_save(cr);
			/* White for highlighted channel, color for active channel */
			if (wwidget->hi_chan == x) {
				cairo_set_source_rgba(cr, 1, 1, 1, 0.20);
			} else {
				cairo_set_source_rgba(cr,
									  wwidget->chancolors[(3 * x) + 0],
									  wwidget->chancolors[(3 * x) + 1],
									  wwidget->chancolors[(3 * x) + 2],
									  0.25);
			}
			cairo_rectangle(cr, start + 0.5, wwidget->g_start_y + 0.5, 
							end - start, wwidget->g_len_y);
			cairo_fill_preserve(cr);
			cairo_stroke(cr);
			cairo_move_to(cr, center, wwidget->g_start_y);
			cairo_line_to(cr, center, wwidget->g_end_y);
			cairo_stroke(cr);
			cairo_restore(cr);

			if (wwidget->hi_chan > -1) {
				snprintf(mtext, 128, "Channel %s, %d%s",
						 wwidget->chanset->chan_text[wwidget->hi_chan],
						 wwidget->chanset->startkhz >= 1000 ? 
						 wwidget->chanset->chan_freqs[wwidget->hi_chan]/ 1000 : 
						 wwidget->chanset->chan_freqs[x],
						 wwidget->chanset->startkhz >= 1000 ? "MHz" : "KHz");

				cairo_save(cr);
				cairo_set_source_rgb(cr, 1, 1, 1);
				cairo_select_font_face(cr, "Helvetica", 
									   CAIRO_FONT_SLANT_NORMAL, 
									   CAIRO_FONT_WEIGHT_BOLD);
				cairo_set_font_size(cr, 14);
				cairo_text_extents(cr, mtext, &extents);
				cairo_move_to(cr, wwidget->g_end_x - extents.width - 5,
							  wwidget->g_start_y + extents.height + 5);
				cairo_show_text(cr, mtext);
				cairo_restore(cr);
			}
		}

	}

	/* Redraw the bounding box */
	cairo_save(cr);
	cairo_rectangle(cr, wwidget->g_start_x, wwidget->g_start_y, 
					wwidget->g_len_x + 0.5, wwidget->g_len_y + 0.5);
	cairo_set_source_rgb(cr, 1, 1, 1);
	cairo_stroke(cr);
	cairo_restore(cr);

	cairo_restore(cr);
}

static gint wispy_widget_expose(GtkWidget *widget, GdkEventExpose *event,
								  gpointer *aux) {
	cairo_t *cr;
	int x, y, w, h;

	g_return_if_fail(widget != NULL);

	cr = gdk_cairo_create(widget->window);

	if (event != NULL) {
		x = event->area.x;
		y = event->area.y;
		w = event->area.width;
		h = event->area.height;
	} else {
		x = 0;
		y = 0;
		w = widget->allocation.width;
		h = widget->allocation.height;
	}

	cairo_rectangle(cr, x, y, w, h);

	cairo_clip(cr);

	wispy_widget_draw(widget, cr, WISPY_WIDGET(aux));

	cairo_destroy(cr);

	return FALSE;
}

void wispy_widget_update(GtkWidget *widget) {
	WispyWidget *wwidget;
	GdkRectangle update_rect;

	g_return_if_fail(widget != NULL);
	g_return_if_fail(IS_WISPY_WIDGET(widget));

	wwidget = WISPY_WIDGET(widget);

	g_return_if_fail(wwidget->draw != NULL);

	update_rect.x = wwidget->draw->allocation.x;
	update_rect.y = wwidget->draw->allocation.y;
	update_rect.width = wwidget->draw->allocation.width;
	update_rect.height = wwidget->draw->allocation.height;

	gtk_widget_draw(widget, &update_rect);

	if (wwidget->update_func != NULL)
		(*(wwidget->update_func))(widget);
}

gint wispy_widget_timeout(gpointer *data) {
	wispy_widget_update(GTK_WIDGET(data));
	return TRUE;
}

/* Annoying that nothing else seems to include this, but we need it for
 * calculating the gradient of colors for the channel highlights */
inline void rgb_to_hsv(double r, double g, double b, 
					   double *h, double *s, double *v) {
	double min, delta;

	if ((b > g) && (b > r)) {
		*v = b;
		if (v != 0) {
			if (r > g)
				min = g;
			else
				min = r;
			delta = *v - min;
			if (delta != 0) {
				*s = (delta / *v);
				*h = 4 + (r - g) / delta;
			} else {
				*s = 0;
				*h = 4 + (r - g);
			}
			*h *= 60;
			if (*h < 0)
				*h += 360;
			*v = *v / 255;
		} else {
			*s = 0;
			*h = 0;
		}
	} else if (g > r) {
		*v = g;
		if (*v != 0) {
			if (r > b)
				min = b;
			else
				min = r;

			delta = *v - min;
			if (delta != 0) {
				*s = (delta / *v);
				*h = 2 + (b - r) / delta;
			} else {
				*s = 0;
				*h = 2 + (b - r);
			}
			*h *= 60;
			if (*h < 0)
				*h += 360;
			*v = *v / 255;
		} else {
			*s = 0;
			*h = 0;
		}
	} else {
		*v = r;
		if (v != 0) {
			if (g > b)
				min = b;
			else
				min = g;

			delta = *v - min;

			if (delta != 0) {
				*s = (delta / *v);
				*h = (g - b) / delta;
			} else {
				*s = 0;
				*h = (g - b);
			}

			*h *= 60;
			if (*h < 0)
				*h += 360;
			*v = *v / 255;
		} else {
			*s = 0;
			*h = 0;
		}
	}
}

inline void hsv_to_rgb(double *r, double *g, double *b, 
					   double h, double s, double v) {
	double hf = h / 60;
	int i = floor(hf);
	double f = hf - i;
	double pv = v * (1 - s);
	double qv = v * (1 - s * f);
	double tv = v * (1 - s * (1 - f));

	if (v == 0) {
		*r = 0;
		*g = 0;
		*b = 0;
		return;
	}

	if (i == -1) {
		*r = v;
		*g = pv;
		*b = qv;
	} else if (i == 0) {
		*r = v;
		*g = tv;
		*b = pv;
	} else if (i == 1) {
		*r = qv;
		*g = v;
		*b = pv;
	} else if (i == 2) {
		*r = pv;
		*g = v;
		*b = tv;
	} else if (i == 3) {
		*r = pv;
		*g = qv;
		*b = v;
	} else if (i == 4) {
		*r = tv;
		*g = pv;
		*b = v;
	} else if (i == 5) {
		*r = v;
		*g = pv;
		*b = qv;
	} else if (i == 6) {
		*r = v;
		*g = tv;
		*b = pv;
	} else {
		*r = 0;
		*b = 0;
		*g = 0;
	}

	*r *= 255;
	*b *= 255;
	*g *= 255;
}

