/* Wispy signal graph
 *
 * GTK widget implementation
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#ifndef __WISPY_GTK_WIDGET_H__
#define __WISPY_GTK_WIDGET_H__

#include "config.h"

#include <glib.h>
#include <glib-object.h>
#include <cairo.h>

#include "wispy_container.h"
#include "wispy_gtk_hw_registry.h"

/* 
 * Common GTK widget layer, provides functions for handling picking a device, 
 * determining and printing channels, etc
 *
 * Should save several hundred lines of shared code off descendant widgets
 * (Spectral, Planar, Topographic, Foo)
 *
 */

G_BEGIN_DECLS

/* Hex color to cairo color */
#define HC2CC(x)				((double) ((double) (x) / (double) 0xFF))

#define WISPY_TYPE_WIDGET \
	(wispy_widget_get_type())
#define WISPY_WIDGET(obj) \
	(G_TYPE_CHECK_INSTANCE_CAST((obj), \
	WISPY_TYPE_WIDGET, WispyWidget))
#define WISPY_WIDGET_CLASS(klass) \
	(G_TYPE_CHECK_CLASS_CAST((klass), \
	WISPY_TYPE_WIDGET, WispyWidgetClass))
#define IS_WISPY_WIDGET(obj) \
	(G_TYPE_CHECK_INSTANCE_TYPE((obj), \
	WISPY_TYPE_WIDGET))
#define IS_WISPY_WIDGET_CLASS(klass) \
	(G_TYPE_CHECK_CLASS_TYPE((class), \
	WISPY_TYPE_WIDGET))

typedef struct _WispyWidget WispyWidget;
typedef struct _WispyWidgetClass WispyWidgetClass;

struct _WispyWidget {
	GtkEventBox parent;

	int hlines;
	/* Top of DB graph, max sample reported */
	int base_db_offset;
	/* Bottom of db graph, min sample reported */
	int min_db_draw;

	/* Graph elements we've calculated */
	int g_start_x, g_start_y, g_end_x, g_end_y,
		g_len_x, g_len_y;
	int dbm_w, chan_h;
	double wbar;

	struct wispy_channels *chanset;
	int *chanhit;
	double *chancolors;

	gint timeout_ref;

	GtkWidget *vbox;
	GtkWidget *sweepinfo;
	GtkWidget *draw, *menubutton;

	wispy_device_registry *wdr;
	wispy_phy *phydev;
	int wdr_slot;

	wispy_sweep_cache *sweepcache;
	GList *wdr_menu;

	/* To be set by children to control behavior */
	int sweep_num_samples;
	int sweep_keep_avg;
	int sweep_keep_peak;
	int sweep_roll_peak;
	int sweep_num_aggregate;
	/* Callbacks used by open, sweep */
	void (* wdr_sweep_func)(int, int, wispy_sample_sweep *, void *);
	void (* wdr_devopen_func)(int, void *);

	/* Graph title (INCLUDING formatting) */
	char *graph_title;

	/* Graph background colors */
	char *graph_title_bg, *graph_control_bg;

	/* Callbacks for drawing */
	void (* draw_func)(GtkWidget *, cairo_t *, WispyWidget *);

	/* Plot channels on the bottom */
	int show_channels;
	/* Show DBM */
	int show_dbm;
	/* Show DBM lines */
	int show_dbm_lines;
	/* Hilighted channel, if we're showing channels */
	int hi_chan;
	/* channel clickable rectangle */
	int chan_start_x, chan_start_y, chan_end_x, chan_end_y;
	GdkPoint *chan_points;

	/* Callbacks for mouse events */
	gint (* draw_mouse_click_func)(GtkWidget *, GdkEventButton *, gpointer *);
	gboolean (* draw_mouse_move_func)(GtkWidget *, GdkEventMotion *, gpointer *);

	/* Update function */
	void (* update_func)(GtkWidget *);
};

struct _WispyWidgetClass {
	GtkEventBoxClass parent_class;
};

GType wispy_widget_get_type(void);
GtkWidget *wispy_widget_new(wispy_device_registry *wdr);
void wispy_widget_clear(void);

/* Do the heavy lifting of actually constructing the GUI, so that child
 * classes can trigger this after setting up titles, etc.  I'm sure this
 * isn't the most elegant method, someone can send me a patch */
void wispy_widget_buildgui(WispyWidget *widget);

/* Calculate the channel clicked in */
inline int wispy_widget_find_chan_pt(WispyWidget *wwidget, int x, int y);

/* Color space conversion tools */
inline void rgb_to_hsv(double r, double g, double b, 
					   double *h, double *s, double *v);
inline void hsv_to_rgb(double *r, double *g, double *b, 
					   double h, double s, double v);

G_END_DECLS

#endif

